@@ -6,12 +6,12 @@ Instructions are in the instructions module.
 import functools
 import string
 import re
+from types import MappingProxyType
 
 from llvmlite.ir import values, types, _utils
 from llvmlite.ir._utils import (_StrCaching, _StringReferenceCaching,
                                 _HasMetadata)
 
-
 _VALID_CHARS = (frozenset(map(ord, string.ascii_letters)) |
                 frozenset(map(ord, string.digits)) |
                 frozenset(map(ord, ' !#$%&\'()*+,-./:;<=>?@[]^_`{|}~')))
@@ -771,7 +771,7 @@ class DIValue(NamedValue):
         return hash((self.is_distinct, self.kind, self.operands))
 
 
-class GlobalValue(NamedValue, _ConstOpMixin):
+class GlobalValue(NamedValue, _ConstOpMixin, _HasMetadata):
     """
     A global value.
     """
@@ -782,6 +782,8 @@ class GlobalValue(NamedValue, _ConstOpMixin):
         super(GlobalValue, self).__init__(*args, **kwargs)
         self.linkage = ''
         self.storage_class = ''
+        self.section = ''
+        self.metadata = {}
 
 
 class GlobalVariable(GlobalValue):
@@ -834,9 +836,15 @@ class GlobalVariable(GlobalValue):
             # emit 'undef' for non-external linkage GV
             buf.append(" " + self.value_type(Undefined).get_reference())
 
+        if self.section:
+            buf.append(", section \"%s\"" % (self.section,))
+
         if self.align is not None:
             buf.append(", align %d" % (self.align,))
 
+        if self.metadata:
+            buf.append(self._stringify_metadata(leading_comma=True))
+
         buf.append("\n")
 
 
@@ -850,19 +858,26 @@ class AttributeSet(set):
     _known = ()
 
     def __init__(self, args=()):
+        super().__init__()
         if isinstance(args, str):
             args = [args]
         for name in args:
             self.add(name)
 
+    def _expand(self, name, typ):
+        return name
+
     def add(self, name):
         if name not in self._known:
             raise ValueError('unknown attr {!r} for {}'.format(name, self))
+        self._validate_add(name)
         return super(AttributeSet, self).add(name)
 
-    def __iter__(self):
-        # In sorted order
-        return iter(sorted(super(AttributeSet, self).__iter__()))
+    def _validate_add(self, name):
+        pass
+
+    def _to_list(self, typ):
+        return [self._expand(i, typ) for i in sorted(self)]
 
 
 class FunctionAttributes(AttributeSet):
@@ -906,18 +921,18 @@ class FunctionAttributes(AttributeSet):
         assert val is None or isinstance(val, GlobalValue)
         self._personality = val
 
-    def __repr__(self):
-        attrs = list(self)
+    def _to_list(self, ret_type):
+        attrs = super()._to_list(ret_type)
         if self.alignstack:
             attrs.append('alignstack({0:d})'.format(self.alignstack))
         if self.personality:
             attrs.append('personality {persty} {persfn}'.format(
                 persty=self.personality.type,
                 persfn=self.personality.get_reference()))
-        return ' '.join(attrs)
+        return attrs
 
 
-class Function(GlobalValue, _HasMetadata):
+class Function(GlobalValue):
     """Represent a LLVM Function but does uses a Module as parent.
     Global Values are stored as a set of dependencies (attribute `depends`).
     """
@@ -934,7 +949,6 @@ class Function(GlobalValue, _HasMetadata):
         self.return_value = ReturnValue(self, ftype.return_type)
         self.parent.add_global(self)
         self.calling_convention = ''
-        self.metadata = {}
 
     @property
     def module(self):
@@ -968,7 +982,8 @@ class Function(GlobalValue, _HasMetadata):
         ret = self.return_value
         args = ", ".join(str(a) for a in self.args)
         name = self.get_reference()
-        attrs = self.attributes
+        attrs = ' ' + ' '.join(self.attributes._to_list(
+            self.ftype.return_type)) if self.attributes else ''
         if any(self.args):
             vararg = ', ...' if self.ftype.var_arg else ''
         else:
@@ -977,9 +992,11 @@ class Function(GlobalValue, _HasMetadata):
         cconv = self.calling_convention
         prefix = " ".join(str(x) for x in [state, linkage, cconv, ret] if x)
         metadata = self._stringify_metadata()
-        pt_str = "{prefix} {name}({args}{vararg}) {attrs}{metadata}\n"
+        metadata = ' {}'.format(metadata) if metadata else ''
+        section = ' section "{}"'.format(self.section) if self.section else ''
+        pt_str = "{prefix} {name}({args}{vararg}){attrs}{section}{metadata}\n"
         prototype = pt_str.format(prefix=prefix, name=name, args=args,
-                                  vararg=vararg, attrs=attrs,
+                                  vararg=vararg, attrs=attrs, section=section,
                                   metadata=metadata)
         buf.append(prototype)
 
@@ -1008,9 +1025,34 @@ class Function(GlobalValue, _HasMetadata):
 
 
 class ArgumentAttributes(AttributeSet):
-    _known = frozenset(['byval', 'inalloca', 'inreg', 'nest', 'noalias',
-                        'nocapture', 'nonnull', 'returned', 'signext',
-                        'sret', 'zeroext'])
+    # List from
+    # https://releases.llvm.org/14.0.0/docs/LangRef.html#parameter-attributes
+    _known = MappingProxyType({
+        # Each tuple is LLVM 11 vs 14 behaviour:
+        # None (unsupported),
+        # True (emit type),
+        # False (emit name only)
+        'byref': (None, True),
+        'byval': (True, True),
+        'elementtype': (None, True),
+        'immarg': (False, False),
+        'inalloca': (False, True),
+        'inreg': (False, False),
+        'nest': (False, False),
+        'noalias': (False, False),
+        'nocapture': (False, False),
+        'nofree': (False, False),
+        'nonnull': (False, False),
+        'noundef': (False, False),
+        'preallocated': (True, True),
+        'returned': (False, False),
+        'signext': (False, False),
+        'sret': (False, True),
+        'swiftasync': (None, False),
+        'swifterror': (False, False),
+        'swiftself': (False, False),
+        'zeroext': (False, False),
+    })
 
     def __init__(self, args=()):
         self._align = 0
@@ -1018,6 +1060,23 @@ class ArgumentAttributes(AttributeSet):
         self._dereferenceable_or_null = 0
         super(ArgumentAttributes, self).__init__(args)
 
+    def _validate_add(self, name):
+        import llvmlite.binding
+        llvm_major = llvmlite.binding.llvm_version_info[0]
+        requires_type = self._known.get(name)[llvm_major >= 14]
+        if requires_type is None:
+            raise ValueError(
+                f"Attribute {name} is not supported on current LLVM version")
+
+    def _expand(self, name, typ):
+        import llvmlite.binding
+        llvm_major = llvmlite.binding.llvm_version_info[0]
+        requires_type = self._known.get(name)[llvm_major >= 14]
+        if requires_type:
+            return f"{name}({typ.pointee})"
+        else:
+            return name
+
     @property
     def align(self):
         return self._align
@@ -1045,8 +1104,8 @@ class ArgumentAttributes(AttributeSet):
         assert isinstance(val, int) and val >= 0
         self._dereferenceable_or_null = val
 
-    def _to_list(self):
-        attrs = sorted(self)
+    def _to_list(self, typ):
+        attrs = super()._to_list(typ)
         if self.align:
             attrs.append('align {0:d}'.format(self.align))
         if self.dereferenceable:
@@ -1078,7 +1137,7 @@ class Argument(_BaseArgument):
     """
 
     def __str__(self):
-        attrs = self.attributes._to_list()
+        attrs = self.attributes._to_list(self.type)
         if attrs:
             return "{0} {1} {2}".format(self.type, ' '.join(attrs),
                                         self.get_reference())
@@ -1092,7 +1151,7 @@ class ReturnValue(_BaseArgument):
     """
 
     def __str__(self):
-        attrs = self.attributes._to_list()
+        attrs = self.attributes._to_list(self.type)
         if attrs:
             return "{0} {1}".format(' '.join(attrs), self.type)
         else:
